


/*++

	Copyright (C) Fur, 2007

:

	pe.h

:

	    PE-

	   :
	ntinside.narod.ru

	        
	


--*/


#include "pe.h"





BOOL
PeLoadImage(
	IN TCHAR *szPeFile,
	IN PE_LOADING_TYPE LoadingType,
	OUT PE_IMAGE* pImage
	)
{
	PIMAGE_DOS_HEADER pDosHeader;
	PIMAGE_NT_HEADERS pNtHeaders;
	
	if ( LoadingType == PeLoadingTypeAsLibrary )
	{
		HMODULE hDll = LoadLibrary( szPeFile );

		if ( hDll )
		{
			pImage->ImageBase = (PVOID) hDll;
			pImage->LoadingType = LoadingType;
		}
		else
			return FALSE;
	}
	else if ( LoadingType == PeLoadingTypeAsBinary )
	{
		HANDLE hFile = CreateFile( 
			szPeFile, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL );

		if ( hFile != INVALID_HANDLE_VALUE )
		{
			DWORD dwSize = GetFileSize( hFile, NULL );

			if ( dwSize > 0 )
			{
				PVOID ImageBase;
				ImageBase = (PVOID) LocalAlloc( LMEM_ZEROINIT, dwSize );

				if ( ImageBase )
				{
					DWORD dwReaded = 0;
					if ( ReadFile( hFile, ImageBase, dwSize, &dwReaded, NULL ) &&
						 (dwReaded == dwSize) )
					{

						pImage->ImageBase = ImageBase;
						pImage->LoadingType = LoadingType;
					}
					else
					{
						LocalFree( pImage->ImageBase );
						CloseHandle( hFile );
						return FALSE;
					}

				}
				else
				{
					CloseHandle( hFile );
					return FALSE;
				}
			}
			else
			{
				CloseHandle( hFile );
				SetLastError( ERROR_INVALID_DATA );
				return FALSE;
			}

			CloseHandle( hFile );
		}
		else
		{
			return FALSE;
		}
	}
	else
	{
		SetLastError( ERROR_UNKNOWN_PROPERTY );
		return FALSE;
	}


	pDosHeader = (PIMAGE_DOS_HEADER) pImage->ImageBase;
	
	if ( pDosHeader->e_magic != IMAGE_DOS_SIGNATURE )
	{
		SetLastError( ERROR_BAD_FORMAT );
		goto CleanupRetFalse;
	}

	pNtHeaders = (PIMAGE_NT_HEADERS) ( (PUCHAR) pImage->ImageBase + pDosHeader->e_lfanew );

	if ( pNtHeaders->Signature != IMAGE_NT_SIGNATURE )
	{
		SetLastError( ERROR_BAD_FORMAT );
		goto CleanupRetFalse;
	}

	//
	//  !!! TODO:   
	//
	if ( pNtHeaders->FileHeader.Machine != IMAGE_FILE_MACHINE_I386 )
	{
		SetLastError( ERROR_EXE_MACHINE_TYPE_MISMATCH );
		goto CleanupRetFalse;
	}

	if ( pNtHeaders->OptionalHeader.Magic != IMAGE_NT_OPTIONAL_HDR32_MAGIC )
	{
		SetLastError( ERROR_BAD_FORMAT );
		goto CleanupRetFalse;
	}

	return TRUE;

CleanupRetFalse:
	PeUnloadImage( pImage );
	
	return FALSE;
}






VOID
PeUnloadImage(
	IN PE_IMAGE* pImage
	)
{
	if ( pImage->LoadingType == PeLoadingTypeAsLibrary )
	{
		FreeLibrary( (HMODULE) pImage->ImageBase );
	}
	else
	{
		LocalFree( pImage->ImageBase );
	}

	pImage->ImageBase = NULL;
}






PVOID
PeGetVA(
	IN PE_IMAGE* pImage,
	IN ULONG RVA 
	)
{

	if ( pImage->LoadingType == PeLoadingTypeAsLibrary )
	{
		//
		//      => VA = ImageBase + RVA
		//
		return (PVOID)( (PUCHAR) pImage->ImageBase + RVA );
	}
	else if ( pImage->LoadingType == PeLoadingTypeAsBinary )
	{
		//
		//         =>    OFFSET
		//
		PIMAGE_DOS_HEADER pDosHeader;
		PIMAGE_NT_HEADERS pNtHeaders;
		PIMAGE_SECTION_HEADER pSection;
		PIMAGE_SECTION_HEADER pSectionWithRVA = NULL;
		WORD i;
		DWORD dwDelta;

		pDosHeader = (PIMAGE_DOS_HEADER) pImage->ImageBase;
		pNtHeaders = (PIMAGE_NT_HEADERS) ( (PUCHAR) pImage->ImageBase + pDosHeader->e_lfanew );

		pSection = IMAGE_FIRST_SECTION( pNtHeaders );

		//
		//   ,     RVA
		//
		for ( i = 0; i < pNtHeaders->FileHeader.NumberOfSections; i++, pSection++ )
		{
			//
			// ,   RVA    
			//
			if ( (RVA >= pSection->VirtualAddress) && 
				 (RVA < (pSection->VirtualAddress + pSection->Misc.VirtualSize)))
			{
				//  
				//    -    
				//  
				pSectionWithRVA = pSection;
				break;
			}
		}

		//
		//     =>  PE
		//
		if ( pSectionWithRVA == NULL )
			return NULL;

		//
		//        RawData  , 
		//  PE   .      0. VA 
		//   VA = ImageBase + RVA - dwDelta
		//
		dwDelta = ( pSectionWithRVA->VirtualAddress - pSectionWithRVA->PointerToRawData);

		return (PVOID)( (PUCHAR) pImage->ImageBase + RVA - dwDelta );
		
	}

	return NULL;
}




PE_EXPORT_INFO*
PeEnumerateExports(
	IN PE_IMAGE *pImage
	)
{
	PE_EXPORT_INFO* pRet = NULL;
	PIMAGE_DOS_HEADER pDosHeader;
	PIMAGE_NT_HEADERS pNtHeaders;
	DWORD i;


	pRet = LocalAlloc( LMEM_ZEROINIT, sizeof(PE_EXPORT_INFO) );

	pDosHeader = (PIMAGE_DOS_HEADER) pImage->ImageBase;
	pNtHeaders = (PIMAGE_NT_HEADERS) ( (PUCHAR) pImage->ImageBase + pDosHeader->e_lfanew );


	//////////////////////////////////////////////////////////////////////////
	//    
	//

	//
	//          
	//
	{
		DWORD *lpdwFunctionsRVAs;
		SHORT *lpdwOrdinals;
		DWORD *lpdwNamesRVAs;
		DWORD *lpdwFunctionsRVAsCopy;
		DWORD dwSymbolIndex;
		CHAR *szLibraryName = NULL;
		PIMAGE_EXPORT_DIRECTORY pExport;
		DWORD dwExportSize;
		PE_EXPORT_SYMBOL *pExportFirst = NULL, *pExportCur = NULL;

		dwExportSize = pNtHeaders->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT].Size;
		
		pExport = (PIMAGE_EXPORT_DIRECTORY) 
			PeGetVA( pImage, pNtHeaders->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT].VirtualAddress );
		

		if ( dwExportSize > 0 )
		{
			lpdwFunctionsRVAs = (DWORD*) 
				PeGetVA( pImage, pExport->AddressOfFunctions );

			lpdwOrdinals = (SHORT*) 
				PeGetVA( pImage, pExport->AddressOfNameOrdinals );

			lpdwNamesRVAs = (DWORD*) 
				PeGetVA( pImage, pExport->AddressOfNames );

			//
			//     -,     RVA .
			//    ,    ,     
			//   ,  RVA    .  RVA - 
			//  ,   ORDINAL
			//
			lpdwFunctionsRVAsCopy = (DWORD*) LocalAlloc( 0, sizeof(DWORD) * pExport->NumberOfFunctions );
			
			RtlCopyMemory( lpdwFunctionsRVAsCopy, lpdwFunctionsRVAs, sizeof(DWORD) * pExport->NumberOfFunctions );

			if ( pExport->Name )
			{
				szLibraryName = (CHAR*) 
					PeGetVA( pImage, pExport->Name );
			}

			//
			//    lpdwNamesRVAs  lpdwOrdinals   .
			//   lpdwOrdinals[i]     lpdwNamesRVAs
			//
			for( i = 0; i < pExport->NumberOfNames; i++ )
			{
				if ( pExportFirst == NULL )
				{
					pExportFirst = (PE_EXPORT_SYMBOL *) LocalAlloc( LMEM_ZEROINIT, sizeof(PE_EXPORT_SYMBOL) );
					pExportCur = pExportFirst;
				}
				else
				{
					pExportCur->pNext = (PE_EXPORT_SYMBOL *) LocalAlloc( LMEM_ZEROINIT, sizeof(PE_EXPORT_SYMBOL) );
					pExportCur = pExportCur->pNext;
				}

				pExportCur->dwRVA = lpdwFunctionsRVAs[i];
				dwSymbolIndex = lpdwOrdinals[i];
				pExportCur->dwOrdinal = dwSymbolIndex + pExport->Base;
				
				if ( dwSymbolIndex < pExport->AddressOfNames )
				{
					CHAR *szName = (CHAR*)
						PeGetVA( pImage, lpdwNamesRVAs[i] );

					pExportCur->szSymbolName = (CHAR*) LocalAlloc( LMEM_ZEROINIT, sizeof(CHAR) * (strlen( szName ) + 1) );
					strncpy( pExportCur->szSymbolName, szName, strlen(szName) );

					pExportCur->dwRVA = lpdwFunctionsRVAsCopy[dwSymbolIndex];
					lpdwFunctionsRVAsCopy[dwSymbolIndex] = 0;
				}
			}

			//
			//    ,     ( ORDINAL)
			//
			for( i = 0; i < pExport->NumberOfFunctions; i++ )
			{
				if ( lpdwFunctionsRVAsCopy[i] != 0 )
				{
					if ( pExportFirst == NULL )
					{
						pExportFirst = (PE_EXPORT_SYMBOL *) LocalAlloc( LMEM_ZEROINIT, sizeof(PE_EXPORT_SYMBOL) );
						pExportCur = pExportFirst;
					}
					else
					{
						pExportCur->pNext = (PE_EXPORT_SYMBOL *) LocalAlloc( LMEM_ZEROINIT, sizeof(PE_EXPORT_SYMBOL) );
						pExportCur = pExportCur->pNext;
					}

					pExportCur->dwOrdinal = i + pExport->Base;
					pExportCur->dwRVA = lpdwFunctionsRVAsCopy[i];
					lpdwFunctionsRVAsCopy[i] = 0;
				}
			}

			LocalFree( lpdwFunctionsRVAsCopy );



			pRet->dwFunctionCount = pExport->NumberOfFunctions;
			pRet->dwFunctionNames = pExport->NumberOfNames;
			pRet->dwOrdinalBase = pExport->Base;
			pRet->dwTimeStamp = pExport->TimeDateStamp;
			
			if ( szLibraryName )
			{
				pRet->szDllName = (CHAR *) LocalAlloc( LMEM_ZEROINIT, sizeof(CHAR) * (strlen(szLibraryName) + 1) );
				strncpy( pRet->szDllName, szLibraryName, strlen( szLibraryName ) );
			}
			
			pRet->pExports = pExportFirst;
		}
		else
		{
			SetLastError( ERROR_NOT_FOUND );
			return FALSE;
		}
	}

	return pRet;
}




VOID
PeFreeExports(
	PE_EXPORT_INFO *pInfo
	)
{
	PE_EXPORT_SYMBOL *pExport, *pNext;

	if ( pInfo->szDllName )
		LocalFree( pInfo->szDllName );

	pExport = pInfo->pExports;

	while ( pExport )
	{
		if ( pExport->szSymbolName )
			LocalFree( pExport->szSymbolName );

		pNext = pExport->pNext;
		LocalFree( pExport );
		pExport = pNext;
	}

}

