/*++

	Copyright (C) Fur, 2007

:

	pe.h

:

	         SDT

	   :
	ntinside.narod.ru

	        
	


--*/



#include "sdt.h"






VOID
SdtFillFunctionInfo(
	IN OUT PSDT_FUNCTION_INFO *ppFirst,
	IN PE_IMAGE *pImage,
	IN PE_EXPORT_SYMBOL *pSymbols,
	IN SDT_SYSTEM_VER SystemVer,
	IN BOOL bExcludeZwFunctions
	)
{
	PE_EXPORT_SYMBOL *pCurSymbol = pSymbols;

	while ( pCurSymbol )
	{
		//
		//       
		//
		PSDT_FUNCTION_INFO pFirst = (*ppFirst);
		PSDT_FUNCTION_INFO pCur = (*ppFirst);
		PSDT_FUNCTION_INFO pLast = NULL;
		DWORD dwNumber = SDT_FUNCTION_NOT_IN_SDT;

		while ( pCur )
		{
			//
			//    ,   
			//
			if ( !stricmp( pCur->szFunctionName, pCurSymbol->szSymbolName ) )
				break;

			pLast = pCur;
			pCur = pCur->pNext;
		}
		
		if ( pCur == NULL )
		{
			//
			//     
			//
			
			if ( pLast == NULL )
			{
				//
				//       
				//

				(*ppFirst) = (PSDT_FUNCTION_INFO) LocalAlloc( LMEM_ZEROINIT, sizeof(SDT_FUNCTION_INFO) );
				pCur = (*ppFirst);	
			}
			else
			{
				//
				//     ,      
				//
				pLast->pNext = (PSDT_FUNCTION_INFO) LocalAlloc( LMEM_ZEROINIT, sizeof(SDT_FUNCTION_INFO) );
				pCur = pLast->pNext;
			}
			
			pCur->szFunctionName = (CHAR*) LocalAlloc( LMEM_ZEROINIT, sizeof(CHAR)*(strlen(pCurSymbol->szSymbolName) + 1 ) );
			strncpy( pCur->szFunctionName, pCurSymbol->szSymbolName, strlen(pCurSymbol->szSymbolName) );

			pCur->dwNumWin2000 = SDT_FUNCTION_NOT_PRESENT;
			pCur->dwNumWinXP = SDT_FUNCTION_NOT_PRESENT;
			pCur->dwNumWin2003 = SDT_FUNCTION_NOT_PRESENT;
			pCur->dwNumWinVista = SDT_FUNCTION_NOT_PRESENT;
		}
		
		//
		// pCur    .        
		//   mov eax, CONST,       
		//  x86    0xB8
		//
		// !!! TODO  

		{
			PVOID pFunctionCode = PeGetVA( pImage, pCurSymbol->dwRVA );
			
			if ( ((PUCHAR) pFunctionCode)[0] == ((UCHAR)0xB8) )
			{
				//
				//       0xB8,   x86 mov eax, const,
				//   const -      SDT
				//
				dwNumber = ((PULONG)((PUCHAR) pFunctionCode + 1))[0];

				//
				//   Rtl      :
				//  mov eax, NTSTATUS
				//  retn STACKSIZE
				//
				//        Nt  Zw

				if ( !((strlen( pCurSymbol->szSymbolName ) > 2) &&
				 (pCurSymbol->szSymbolName[0] == 'Z') &&
				 (pCurSymbol->szSymbolName[1] == 'w')) &&
					
					!((strlen( pCurSymbol->szSymbolName ) > 2) &&
				 (pCurSymbol->szSymbolName[0] == 'N') &&
				 (pCurSymbol->szSymbolName[1] == 't')) )
					dwNumber = SDT_FUNCTION_NOT_IN_SDT;
				 
			}
		}

		if ( bExcludeZwFunctions )
		{
			if ( (strlen( pCurSymbol->szSymbolName ) > 2) &&
				 (pCurSymbol->szSymbolName[0] == 'Z') &&
				 (pCurSymbol->szSymbolName[1] == 'w') )
				 dwNumber = SDT_FUNCTION_NOT_PRESENT;
		}
		
		switch( SystemVer )
		{
		case SdtSystemWindows2000:
			pCur->dwNumWin2000 = dwNumber;
			break;

		case SdtSystemWindowsXP:
			pCur->dwNumWinXP = dwNumber;
			break;

		case SdtSystemWindows2003:
			pCur->dwNumWin2003 = dwNumber;
			break;

		case SdtSystemWindowsVista:
			pCur->dwNumWinVista = dwNumber;
			break;
		}


		pCurSymbol = pCurSymbol->pNext;
	}
}







VOID
SdtFreeFunctionInfo(
	IN PSDT_FUNCTION_INFO pInfo
	)
{
	PSDT_FUNCTION_INFO pCur, pNext;

	pCur = pInfo;

	while( pCur )
	{
		LocalFree( pCur->szFunctionName );
		pNext = pCur->pNext;
		LocalFree( pCur );
		pCur = pNext;
	}
}





BOOL
SdtProcessDll(
	IN OUT PSDT_FUNCTION_INFO *ppFirst,
	IN TCHAR *szDllFile,
	IN SDT_SYSTEM_VER SystemVer
	)
{
	BOOL bRet;
	PE_EXPORT_INFO *pInfo;
	PE_IMAGE Image;

	bRet = PeLoadImage( szDllFile, 
		PeLoadingTypeAsLibrary, 
		&Image );

	if ( bRet )
	{
		pInfo = PeEnumerateExports( &Image );

		if ( pInfo )
		{
			SdtFillFunctionInfo( ppFirst, &Image, pInfo->pExports, SystemVer, TRUE );

			PeFreeExports( pInfo );
		}
		else
			bRet = FALSE;

		PeUnloadImage( &Image );
	}
	else
		bRet = FALSE;

	return bRet;
}




PSDT_FUNCTION_INFO
SdtInspectDlls(
	IN TCHAR *szDllWindows2000,
	IN TCHAR *szDllWindowsXP,
	IN TCHAR *szDllWindows2003,
	IN TCHAR *szDllWindowsVista
	)
{
	PSDT_FUNCTION_INFO pRet = NULL;

	if ( szDllWindows2000 )
		if ( !SdtProcessDll( &pRet, szDllWindows2000, SdtSystemWindows2000 ) )
		{
			_ftprintf( stderr, _T("-Can't process DLL for Windows 2000\n") );
		}

	if ( szDllWindowsXP )
		if ( !SdtProcessDll( &pRet, szDllWindowsXP, SdtSystemWindowsXP ) )
		{
			_ftprintf( stderr, _T("-Can't process DLL for Windows XP\n") );
		}

	if ( szDllWindows2003 )
		if ( !SdtProcessDll( &pRet, szDllWindows2003, SdtSystemWindows2003 ) )
		{
			_ftprintf( stderr, _T("-Can't process DLL for Windows 2003\n") );
		}

	if ( szDllWindowsVista )
		if ( !SdtProcessDll( &pRet, szDllWindowsVista, SdtSystemWindowsVista ) )
		{
			_ftprintf( stderr, _T("-Can't process DLL for Windows Vista\n") );
		}

	return pRet;
}



VOID
SdtPrintTable(
	IN PSDT_FUNCTION_INFO pInfo,
	IN BOOL bHexNumbers
	)
{
	PSDT_FUNCTION_INFO pCur = pInfo;

	_tprintf( _T("%-40s%-10s%-10s%-10s%-10s\n"), _T("Function"), _T("2000"), _T("XP"), _T("2003"), _T("Vista") );
	_tprintf( _T("\n") );

	while ( pCur )
	{
		if ( (pCur->dwNumWin2000 < SDT_FUNCTION_NOT_PRESENT) ||
			 (pCur->dwNumWinXP < SDT_FUNCTION_NOT_PRESENT) ||
			 (pCur->dwNumWin2003 < SDT_FUNCTION_NOT_PRESENT) ||
			 (pCur->dwNumWinVista < SDT_FUNCTION_NOT_PRESENT)
			 )
		{
			TCHAR szWin2000Num[100];
			TCHAR szWinXPNum[100];
			TCHAR szWin2003Num[100];
			TCHAR szWinVistaNum[100];


			if ( pCur->dwNumWin2000 >= SDT_FUNCTION_NOT_PRESENT )
				_tcscpy( szWin2000Num, _T("-") );
			else
			{
				if ( bHexNumbers )
					_stprintf( szWin2000Num, _T("0x%.6X"), pCur->dwNumWin2000 );
				else
					_stprintf( szWin2000Num, _T("%d"), pCur->dwNumWin2000 );
			}

			if ( pCur->dwNumWinXP >= SDT_FUNCTION_NOT_PRESENT )
				_tcscpy( szWinXPNum, _T("-") );
			else
			{
				if ( bHexNumbers )
					_stprintf( szWinXPNum, _T("0x%.6X"), pCur->dwNumWinXP );
				else
					_stprintf( szWinXPNum, _T("%d"), pCur->dwNumWinXP );
			}


			if ( pCur->dwNumWin2003 >= SDT_FUNCTION_NOT_PRESENT )
				_tcscpy( szWin2003Num, _T("-") );
			else
			{
				if ( bHexNumbers )
					_stprintf( szWin2003Num, _T("0x%.6X"), pCur->dwNumWin2003 );
				else
					_stprintf( szWin2003Num, _T("%d"), pCur->dwNumWin2003 );
			}


			if ( pCur->dwNumWinVista >= SDT_FUNCTION_NOT_PRESENT )
				_tcscpy( szWinVistaNum, _T("-") );
			else
			{
				if ( bHexNumbers )
					_stprintf( szWinVistaNum, _T("0x%.6X"), pCur->dwNumWinVista );
				else
					_stprintf( szWinVistaNum, _T("%d"), pCur->dwNumWinVista );
			}


			if ( strlen( pCur->szFunctionName ) > 40 )
			{
				_tprintf( _T("%s\n"), pCur->szFunctionName );

				_tprintf( _T("%-40s%-10s%-10s%-10s%-10s\n"), 
					_T(""), szWin2000Num, szWinXPNum, szWin2003Num, szWinVistaNum );
			}
			else
			{
				_tprintf( _T("%-40s%-10s%-10s%-10s%-10s\n"), 
					pCur->szFunctionName, szWin2000Num, szWinXPNum, szWin2003Num, szWinVistaNum );
			}

		}

		pCur = pCur->pNext;
	}
}



VOID
SdtUsage()
{
	_tprintf("\
SDT Tool. Copyright (C) Fur, 2007\n\
This tool build the table of functions' SDT-numbers\n\
for Windows 2000, Windows XP, Windows 2003, Windows Vista\n\
\n\
Tool works only with x86 ntdll.dll files from this systems\n\
\n\
Usage: sdt [/hex] [/win2000:path_to_ntdll]\n\
  [/winxp:path_to_ntdll] [/win2003:path_to_ntdll] [/winvista:path_to_ntdll]\n\
\n\
Example:\n\
sdt /winxp:C:\\Windows\\System32\\ntdll.dll /win2000:C:\\win2000files\\ntdll.dll\n\
");

}



_tmain(
	INT argc,
	CHAR *argv[]
	)
{
	
	if ( argc <= 1 )
	{
		SdtUsage();
		return -1;
	}
	else
	{
		TCHAR *szWin2000 = _T("/win2000");
		TCHAR *szWinXP = _T("/winxp");
		TCHAR *szWin2003 = _T("/win2003");
		TCHAR *szWinVista = _T("/winvista");
		TCHAR *szHex = _T("/hex");
		INT i;

		BOOL bHex = FALSE;
		TCHAR *szWin2000Dll = NULL;
		TCHAR *szWinXPDll = NULL;
		TCHAR *szWin2003Dll = NULL;
		TCHAR *szWinVistaDll = NULL;

		PSDT_FUNCTION_INFO pInfo;

		for( i = 0; i < argc; i++ )
		{
			if ( !_tcsicmp( argv[i], szHex ) )
				bHex = TRUE;

			if ( !_tcsnicmp( argv[i], szWin2000, _tcslen(szWin2000) ) )
				szWin2000Dll = argv[i] + _tcslen(szWin2000) + 1;

			if ( !_tcsnicmp( argv[i], szWinXP, _tcslen(szWinXP) ) )
				szWinXPDll = argv[i] + _tcslen(szWinXP) + 1;

			if ( !_tcsnicmp( argv[i], szWin2003, _tcslen(szWin2003) ) )
				szWin2003Dll = argv[i] + _tcslen(szWin2003) + 1;

			if ( !_tcsnicmp( argv[i], szWinVista, _tcslen(szWinVista) ) )
				szWinVistaDll = argv[i] + _tcslen(szWinVista) + 1;
		}

		if ( (szWin2000Dll == NULL) && 
			(szWinXPDll == NULL) && 
			(szWin2003Dll == NULL) &&
			(szWinVistaDll == NULL) )
		{
			_ftprintf( stderr, _T("-No ntdll libraries specified\n") );
		}
		else
		{
			pInfo = SdtInspectDlls( szWin2000Dll, szWinXPDll, szWin2003Dll, szWinVistaDll );
			SdtPrintTable( pInfo, bHex );
			SdtFreeFunctionInfo( pInfo );
		}

		return 1;
	}

	return 0;
}


